﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace FileCRC
{
    class Program
    {
        static void Main(string[] args)
        {
            try
            {
                string name = "TestIn.bin";
                // check file
                if (!File.Exists(name))
                    throw new Exception(name + " not found");
                // read file
                byte[] data = File.ReadAllBytes(name);
                // check CRC
                vRSA rsa = new vRSA((ulong)System.Environment.TickCount);
                if (rsa.CheckKey(data) == false)
                    throw new Exception("Incorrect key");
                // remove CRC
                byte[] buf = new byte[data.Length - 8];
                Array.Copy(data, buf, data.Length - 8);
                // calculate CRC
                ulong key = rsa.ComputePrivateKey(buf);
                // add CRC
                BitConverter.GetBytes(key).CopyTo(data, buf.Length);
                // write file
                File.WriteAllBytes("TestOut.bin", data);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
                Console.WriteLine("Press any key...");
                Console.ReadKey();
            }
        }
    }

    internal class vRSA
    {
        private ulong P, Q, M, N, /*E, */Y;

        public vRSA(ulong p)
        {
            P = p ^ 0x151D8255;
            Q = p ^ 0x274ECC00;
            M = 0x67789ED4559AF79;
            N = 0xCCCCCCCCCCCCCCCC;
            //E = 0x5D405B5;
            Y = 0x53290744C4D541;
        }

        public ulong ComputeKey(byte[] data)
        {
            return GetKey(PrepareKey(data) % N, N, M);
        }

        // NOT WORK!!!
        public ulong ComputePrivateKey(byte[] data)
        {
            ulong dataKey = PrepareKey(data);
            ulong keyDt = dataKey % M;
            return GetKey(keyDt, Y, M);
        }

        public bool CheckKey(byte[] data)
        {
            if (data.Length < 8)
                return false;
            int szData = data.Length - 8;
            byte[] buf = new byte[szData];
            Array.Copy(data, buf, szData);
            ulong dataKey = PrepareKey(buf);
            ulong origKey = BitConverter.ToUInt64(data, szData);
            ulong keyOr = GetKey(origKey, Y, M);
            ulong keyDt = dataKey % M;
            return keyOr == keyDt;
        }

        private ulong PrepareKey(byte[] data)
        {
            if (data.Length < 1)
                return 0;
            ulong h = 0;
            ulong pm = 0x123456789;
            for (int i = 0; i < data.Length / 8; i++)
            {
                ulong w = BitConverter.ToUInt64(data, i * 8);
                h ^= w;
                uint lw = (uint)w;
                uint hw = (uint)(w >> 32);
                ulong sign = ((hw & 0x80000000) != 0) ? 0xFFFFFFFF00000000 : 0;
                ulong t = ((ulong)(hw ^ lw) << 32) | (ulong)(lw ^ hw);
                t += sign;
                t += (((ulong)lw << 32) | hw) | sign;
                t += pm;
                t += w;
                pm ^= t;
            }
            int rem = data.Length & 7;
            if (rem != 0)
            {
                long ls = 0;
                long rs = 0;
                long mdc = 0x11F71FB04CB;
                int last = 0;
                if (rem >= 2)
                {
                    int it = (rem - 2) / 2 + 1;
                    last = it * 2;
                    int off = data.Length - rem;
                    int sh = 0;
                    for (int i = 0; i < it; i++, sh += 16)
                    {
                        ls += (long)((sbyte)data[i * 2 + off]) << sh;
                        rs += (long)((sbyte)data[i * 2 + 1 + off]) << (sh + 8);
                    }
                }
                if (last < rem)
                    mdc += (long)((sbyte)data[data.Length - 1]) << (last << 8);
                ulong w = (ulong)(ls + rs + mdc);
                h ^= w;
                uint lw = (uint)w;
                uint hw = (uint)(w >> 32);
                ulong sign = ((hw & 0x80000000) != 0) ? 0xFFFFFFFF00000000 : 0;
                ulong t = ((ulong)(hw ^ lw) << 32) | (ulong)(lw ^ hw);
                t += sign;
                t += (((ulong)lw << 32) | hw) | sign;
                t += pm;
                t += w;
                pm ^= t;
            }
            return (((ulong)data.Length * 0x100000001) ^ pm ^ h) & 0xFFFFFFFFFFF;
        }

        private ulong GetKey(ulong rem, ulong n, ulong m)
        {
            ulong key = 1;
            ulong prv = rem;
            for (int i = 0; i < 64; i++)
            {
                if (((n >> i) & 1) != 0)
                    key = UpdateKey(key, prv, m);
                prv = UpdateKey(prv, prv, m);
            }
            return key;
        }

        private ulong UpdateKey(ulong k, ulong n, ulong m)
        {
            ulong key = 0;
            ulong prv = k;
            for (int i = 0; i < 64; i++)
            {
                if (((n >> i) & 1) != 0)
                    key = (key + prv) % m;
                prv = prv * 2 % m;
            }
            return key;
        }
    }
}
